.LISTMAC

.NOLIST
.INCLUDE "m32def.inc"
.LIST

.DEF xlsi		= R0		;XI(Interrupt) Save
.DEF xhsi		= R1
.DEF ticks		= R2		;15625Hz Ticker

.DEF NULL		= R5
.DEF FULL		= R6
.DEF ssreg		= R7		;Status Reg Save
.DEF ibupw		= R8		;Input Buffer Pointer Write
.DEF ibupr		= R9		;dito Read

.DEF t0rlv		= R11
.DEF spipbein	= R12
.DEF spipbaus	= R13
.DEF spiein		= r14
.DEF spiaus		= r15

.DEF mpa		= R16		;Akku	A
.DEF mpb		= R17		;		B
.DEF mpc		= R18		;		C
.DEF zz			= R19		;Zeilenzhler
.DEF fz			= R20		;Framezhler

.DEF cgp		= R24
.DEF outdata	= R25		;Ausgaberegister

.EQU timval		= 127		;Timerout PORTA, FULL
.EQU txtl		= 25		;Anzahl Text Lines
.EQU preql		= 2			;Pre Vsync Equalizing Lines
.EQU vsyncl		= 3			;Vertical Sync Lines
.EQU poeql		= 2			;Post Vsync Equalizing Lines
.EQU vidl		= txtl		;Video Lines
.EQU ubll		= 50		;upper blank lines
.EQU lbll		= 30		;lower blank lines
.EQU cpl		= 40		;Chars per line
.EQU ssv		= 56		;Schwarzschulter vorne
.EQU lpc		= 9			;lines per Char
.EQU CBSize		= cpl*vidl	;Charbufsize
.EQU InBufSize	= 256		;ser.InputBuffer
.EQU DevType 	= 0x20

.DSEG
Charbuf:	.byte CBSize		;ASCII Zeichenspeicher
inbuf:		.byte inbufsize  	;Eingangsbuffer fr RSR232
cur:		.byte 1				;Cursorzeichen
curpl:		.byte 1
curph:		.byte 1
curbuf:		.byte 1
scroll:		.byte 1
clrs:		.byte 1
posc:		.byte 1
.CSEG

.macro endint
;Ende der Interruptverarbeitung
;damit werden die Zeilen-INTs beendet
;muss vor dem nchsten Timer-Int erreicht werden 
	out SREG,ssreg				;Restore StatusReg
	reti						;und ab nach Hause
.endmacro

.macro synclo
	cbi portb,4
.endmacro

.macro synchi
	sbi portb,4
.endmacro


;Interrupt Vector Table
.ORG 0x0000
	jmp RESET		; Reset Handler
	jmp RESET		; IRQ0 Handler
	jmp RESET		; IRQ1 Handler
	jmp RESET		; IRQ2 Handler
	jmp RESET		; Timer2 Compare Handler
	jmp RESET		; Timer2 Overflow Handler
	jmp RESET		; Timer1 Capture Handler
	jmp RESET		; Timer1 CompareA Handler
	jmp RESET		; Timer1 CompareB Handler
	jmp RESET		; Timer1 Overflow Handler
	jmp RESET		; Timer0 Compare Handler
	rjmp hsync		; Timer0 Overflow Handler
	nop
	jmp RESET		; SPI Transfer Complete Handler
	jmp RESET		; USART RX Complete Handler
	jmp RESET		; UDR Empty Handler
	jmp RESET		; USART TX Complete Handler
	jmp RESET		; ADC Conversion Complete Handler
	jmp RESET		; EEPROM Ready Handler
	jmp RESET		; Analog Comparator Handler
	jmp RESET		; Two-wire Serial Interface Handler
	jmp RESET		; Store Program Memory Ready Handler


RESET:  
;SP setzen
	ldi mpa,high(RAMEND)
	out SPH,mpa
	ldi mpa,low(RAMEND)
	out SPL,mpa
;Init IO-Ports
	ldi mpa,0x00
	mov NULL, mpa
	ldi mpa, 0xFF
	mov FULL, mpa
	ldi mpa,0b10110000
	mov spipbein,mpa
	ldi mpa,0b10010010
	mov spipbaus,mpa
	ldi mpa,0x50
	mov spiein,mpa
	ldi mpa,0x10			;0x10 geht auch
	mov spiaus,mpa
	out ddrb,spipbaus		;PortB0 Output, B1-B7 Input
	ldi mpa,0x00
	out portb,mpa
	cbi portb,4				;Video Out auf LO, Sync Out bleibt HI
;SPI auf double Clk
	ldi mpa,0x01
	out spsr,mpa

;Zeichenspeicher mit Leerzeichen fllen
	ldi XH,HIGH(Charbuf)
	ldi XL, LOW(Charbuf)
	ldi mpc,' '
	ldi mpa,txtl
ldtst1:
	ldi mpb,cpl
ldtst2:
	st X+,mpc
	dec mpb
	brne ldtst2
	dec mpa
	brne ldtst1
  
;Starttext eintragen
	ldi ZH,HIGH(Initxt<<1)
	ldi ZL, LOW(Initxt<<1)
	ldi XH,HIGH(Charbuf)
	ldi XL, LOW(Charbuf)
	ldi mpb,HIGH(cpl*20)	;19 Zeilen
	inc mpb
	ldi mpa, LOW(cpl*20)
inil1:
	lpm mpc,Z+
	st X+,mpc
	dec mpa
	brne inil1
	ldi mpa, 0xFF
	dec mpb
	brne inil1
	
;Input Buffer lschen
	ldi XH,HIGH(inbuf)
	ldi XL, LOW(inbuf)
	ldi mpa,InBufSize-1
	ldi mpb, 0x00
inbufc:
	st X+,mpb
	subi mpa, 1
	brcc inbufc

;Init TIMER0
	ldi mpa,0x02
	out tccr0,mpa			;timer 0 vorteiler auf clk/8
	ldi mpa,0-timval		;timer0 reload value 63us
	mov t0rlv,mpa			;wird noch fter gebraucht
	out tcnt0,t0rlv
	ldi mpa,0x01
	out timsk,mpa			;timer 0 interrupt enable

;Register vorbelegen / lschen
	clr ibupr
	clr ibupw
	ldi mpa,' '
	sts curbuf,mpa
	ldi XH,HIGH(Charbuf+cpl*20) ;Cursor auf Zeile 19
	ldi XL, LOW(Charbuf+cpl*20)
	sts curph,XH
	sts curpl,XL		
	sts scroll, NULL			;kein Scrollen
	sts clrs, NULL				;kein Cls
	clr mpa					;keine CursorPos Sequenz aktiv
	sbr mpa, 4					;Cursorshift aktiv
	sbr mpa, 8					;Cursor aktiv
	sts posc, mpa
	ldi mpa,'_'
	sts cur,mpa					;Cursor setzen
	ldi ZH,HIGH(preq)			;Starten mit VSync-Zeilen
	ldi ZL, LOW(preq)
	ldi zz,preql				;Anzahl Zeilen
	sei							;Interupt Enable

;Das Hauptprogamm
;Okay, ist nicht viel, aber schn bersichtlich
mloop:
	nop
	rjmp mloop
;Ende Hauptprogramm 
;was sonst noch zu tun ist, luft im Interrupt 

;Baudtab & Initext
.include "tabs_32.txt"

;Horizontal-Sync, damit startet jede Zeile
;wird alle 64us per Timer0-Int aufgerufen
;erzeugt das H-Sync Signal
;Startet den 64us Timer neu
;Holt, falls ntig, ein neues Zeichen vom Buffer
;und legt es im Input-Buffer ab
hsync:
	nop				;Laufzeitausgleich
	synclo				;Syncsignal Lo setzen
	out tcnt0,t0rlv
	in ssreg,SREG		;Statusreg sichern
	mov xhsi,XH			;X sichern
	mov xlsi,XL

	ldi mpb,22			;normale Hsynczeit
	nop
	sbic PINC, 0		;CS prfen
	rjmp hsync3			;kein CS (High) dann Ende
	nop							;Ausgleichszeit
	nop
	nop
	ldi mpa, 0x18		;ACK und ERR als Ausgang definieren
	out PORTC, mpa		;und Heigh setzen
	out DDRC, mpa
	subi mpb, 4			;Zeit anpassen
	sbis PINC, 2		;WR prfen
	rjmp write			;wenn Low dann Springen
	sbis PINC, 5		;TYP prfen
	rjmp type
	sbis PINC, 1		;RD prfen
	rjmp read
	nop
	inc mpb					;Zeit anpassen
hsync1:
	mov XH,xhsi			;Restore X
	mov XL,xlsi
hsync2:
	dec mpb				;Restliche HSync-Zeit vertrdeln
	brne hsync2
	ijmp				;Ende Hsync
hsync3:
	out PORTC, NULL		;Wenn kein CS dann
	out DDRC, NULL		;alles Hochohmig
	nop
	nop
	rjmp hsync1

read:					;Von Device lesen
	out DDRD, FULL
	out PORTD, outdata	;Daten setzen
	cbi PORTC, 3		;ACK auf Low
reada:
	subi mpb, 2			;Warten bis RD High wird
	breq exit			;oder die Zeit abgelaufen ist
	nop
	sbis PINC, 1
	rjmp reada
	sbi PORTC, 3		;Wenn RD High ist dann ACK auf High
	out DDRD, NULL		;PORTD hochohmig
	out PORTD, NULL
	subi mpb, 1			;Zeit anpassen
	mov XH, xhsi		;X wiederherstellen
	mov XL, xlsi
readl:
	dec mpb				;Restliche Zeit vertdeln
	brne readl
	ijmp

write:
	ldi XH,HIGH(inbuf)	;X auf Start Input Buffer
	ldi XL, LOW(inbuf)
	add XL,ibupw		;Write-Pointer dazu
	adc XH, NULL
	in mpa, PIND		;Daten vom Bus lesen
	cbi PORTC, 3		;ACK auf Low
writea:
	subi mpb, 2			;Warten bis WR High wird
	breq exit			;oder die Zeit abgelaufen ist
	nop
	sbis PINC, 2
	rjmp writea
	sbi PORTC, 3 		;Wenn WR High ist dann ACK auf High
	st X,mpa			;neuen Char ablegen
	inc ibupw			;Pointer weitersetzen
	ldi mpa,InBufSize-1	;Pointer wraparound
	and ibupw,mpa
	mov XH, xhsi		;X wiederherstellen
	mov XL, xlsi
	subi mpb, 2			;Zeitausgleich
	breq writee
writel:
	dec mpb				;Zeit vertrdeln
	brne writel
	ijmp
writee:					;...oder gleich Beenden
	nop
	ijmp
	
type:
	ldi mpc, DevType
	out DDRD, FULL
	out PORTD, mpc	;Daten setzen
	cbi PORTC, 3		;ACK auf Low
typea:
	subi mpb, 2			;Warten bis RD High wird
	breq exit			;oder die Zeit abgelaufen ist
	nop
	sbis PINC, 5
	rjmp typea
	sbi PORTC, 3		;Wenn RD High ist dann ACK auf High
	out DDRD, NULL		;PORTD hochohmig
	out PORTD, NULL
	subi mpb, 1			;Zeit anpassen
	mov XH, xhsi		;X wiederherstellen
	mov XL, xlsi
typel:
	dec mpb				;Restliche Zeit vertdeln
	brne typel
	ijmp

exit:					;Wenn die Zeit abgelaufen ist
	sbi PORTC, 3		;ACK auf High
	out DDRD, NULL		;PORTD hochohmig
	out PORTD, NULL
	cbi PORTC, 4		;ERR auf Low
	nop
	mov XH,xhsi			;Restore X
	mov XL,xlsi
	ijmp

;Leerzeilen mit den Ausgleichsimpulsen
preq:
	synchi				;Syncsignal wieder auf HI
	ldi mpa,145			;1/2 Zeile abwarten
preqb:
	dec mpa 
	brne preqb
	;Hsync Impuls auf 1/2 Zeile
	cbi portb,4			;Sync auf LO
	ldi mpa,24
preqc:
	dec mpa
	brne preqc
	sbi portb,4			;Sync zuck auf HI

	dec zz				;noch mehr Zeilen ?
	brne preqd			;ja
	ldi ZH,HIGH(vsync)	;danach kommt Vsync
	ldi ZL,LOW(vsync)
	ldi zz,vsyncl		;Anzahl setzen
preqd:
	endint				;das wars

;Vertical Sync
;damit fngt jedes Bild an
vsync: ;1/2 Zeile minus HSync-Zeit abwarten, Sync bleibt noch LO
	ldi mpa,120 
vs1a:
	dec mpa
	brne vs1a
	;inv. Hsync-Impuls auf 1/2 Vsync-Zeile
	sbi portb,4			;Sync auf HI
	ldi mpa,24			;4,75us
vs1b:
	dec mpa
	brne vs1b
	synclo				;Sync auf LO
	ldi mpa,142			;  fr den Rest der Zeile
vs1c:
	dec mpa
	brne vs1c
	synchi				;Sync zurck auf HI
	dec zz				;noch mehr Vsync-Zeilen ?
	brne vs1d			;  ja
	ldi ZH,HIGH(poeq)	;danach kommen die
	ldi ZL,LOW(poeq)	;  hinteren Ausgleichs-Zeilen
	ldi zz,poeql		;Anzahl setzen
vs1d:
	endint

;Leerzeilen mit den Ausgleichsimpulsen
poeq:
	synchi				;Syncsignal wieder auf HI
	ldi mpa,144			;1/2 Zeile abwarten
poeqb:
	dec mpa 
	brne poeqb
	;Hsync Impuls auf 1/2 Zeile
	synclo				;Sync auf LO
	ldi mpa,24
poeqc:
	dec mpa
	brne poeqc
	synchi				;Sync zuck auf HI

	dec zz				;noch mehr Zeilen ?
	brne poeqd			; ja
	ldi ZH,HIGH(leero)	;danach kommen Leerzeilen
	ldi ZL,LOW(leero)
	ldi zz,ubll			;Anzahl setzen
poeqd:
	endint				;das wars

;Obere Leerzeilen
;Hier ist der Terminalprogrammteil mit
;untergebracht, weil hier viel Zeit brig ist
leero:
	synchi				;Syncsignal wieder auf HI
	rcall term			;Terminalteil ausfhren 

leeroa:
	dec zz				;noch mehr Leerzeilen ?
	brne leerob			;  ja
	ldi ZH,HIGH(dline)	;als nchstes kommen 
	ldi ZL, LOW(dline)	;Display Lines
	ldi zz,lpc			;Zeilen pro Zeichen
	ldi fz,vidl			;Anzahl Video-Zeilen
	ldi YH,HIGH(Charbuf);Init Character-Bufferpointer
	ldi YL, LOW(Charbuf)
	ldi cgp,High(chargenx<<1)
leerob:
	endint 

;Hier ist die eigentliche Video-Darstellung
dline:
	synchi				;Syncsignal wieder auf HI
	;Schwarzschulter abwarten
	ldi mpa,ssv
dlineb:
	dec mpa
	brne dlineb
	;Pixeldaten fr 40 Chars rausschieben
	mov XH,YH
	mov XL,YL
	mov ZH,cgp
	inc cgp
	ldi mpc,cpl
dlinelp:
	ld ZL,x+
	lpm mpb,Z
	out ddrb,spipbaus
	out spcr,spiaus
	out spcr,spiein 
	out spdr,mpb
	nop
	out ddrb,spipbein
	andi mpb,0x02
	ori mpb, 0x10
	out portb,mpb
	dec mpc
	brne dlinelp
	nop
	nop
	nop
	nop
	ldi mpa, 0x10
	out portb,mpa  
	out spcr,spiaus
	out ddrb,spipbaus

	dec zz				;alle Zeilen fertig ?
	brne dlined		
	;  nee
	dec fz				;waren das alle Frames ?
	brne dlinec			;nee, da kommen noch mehr
	ldi ZH,HIGH(leeru)	;Ok, dann machen wir weiter
	ldi ZL, LOW(leeru)	;mit den unteren Leerzeilen
	ldi zz,lbll			;Anzahl Leerzeilen
	endint				;Zurck
	;noch ein Frame, also vorher neue Pixel bereitstellen
dlinec:
	ldi ZH,HIGH(dline)	;weitere video lines
	ldi ZL, LOW(dline)
	ldi zz,lpc			;Zeilen pro Zeichen
	adiw YH:YL,cpl		;Chars per line
	ldi cgp,High(chargenx<<1)
	endint
dlined:
	ldi ZH,HIGH(dline)	;restore wg.
	ldi ZL, LOW(dline)	;Wiederholung letzte Zeile
	cpi zz,1
	brne dlinee
	dec cgp
dlinee:
	endint

;Untere Leerzeilen
leeru:
	synchi				;Syncsignal wieder auf HI
	rcall term			;Terminalteil ausfhren
	dec zz 				;noch mehr Zeilen ?
	brne leerua			;ja
	ldi ZH,HIGH(preq)	;Bild Ende
	ldi ZL,LOW(preq)	;jetzt kommt der nchste VSync
	ldi zz,preql		;Anzahl Vsync Zeilen
	inc ticks			;50Hz Ticker
leerua:
	endint

charwx:					;Zwischenstation
	rjmp charw

;Der Terminalprogrammteil
term:
	lds mpc, clrs
	ori mpc,0				;sind wir am Bildschirm-Lschen ?
	brne clrscr				;  ja
	lds mpc, scroll
	cpi mpc,1			;sind wir am Scrollen ?
	brlo charwx				;  nee
	;also scrollen
	breq scrlpb				;letzte Scrollzeile, also neue Leerzeile
	lds XH,curph			;akt. Cursorposition im Charbuffer
	lds XL,curpl
	lds mpa, curbuf
	st X,mpa				;Restore Char auf Cursorposition
	dec mpc				;nchste Scroll-Portion
	ldi mpa,txtl			;gescrollt wird von unten nach oben
	sub mpa,mpc
	ldi mpb,cpl				;eine Zeile scrollen
	mul mpb,mpa				;akt. Portion mal Portionsgrsse
							;= Scroll-Startposition im Buffer
							
	sts scroll, mpc
	ldi YH,HIGH(Charbuf)	;Anfang Buffer
	ldi YL, LOW(Charbuf)  
	ldi XH,HIGH(Charbuf+cpl);Anfang zweite Zeile
	ldi XL, LOW(Charbuf+cpl)
	add YL,r0				;plus Scrollstartposition
	adc YH,r1
	add XL,r0				;auch plus Scrollstart
	adc XH,r1
scrlp:
	ld mpa,X+
	st Y+,mpa
	dec mpb
	brne scrlp				;fertig
scrlpa:
	lds XH,curph			;X wieder auf akt. Cursorpos.
	lds XL,curpl
	;Cursor blinken lassen
	lds mpb, posc
	sbrc mpb, 3			;nur blinken lassen falls aktiv
	lds mpa,cur
	sbrs mpb, 3
	lds mpa,curbuf
	sbrc ticks,4			;ca. 2Hz
	lds mpa,curbuf
	st X,mpa
	ret						;zurck zum INT-Handler
	;letzte Zeile auf Bildschirm lschen
scrlpb:
	sts scroll, NULL				;Scrollen ist fertig
	ldi YH,HIGH(Charbuf+cbsize-cpl)	;Pointer auf letzte Zeile
	ldi YL, LOW(Charbuf+cbsize-cpl)	;im Character Buffer 
	ldi mpb,cpl						;1 Zeile lschen
	ldi mpa,' '						;Lschen = Leerzeichen eintragen 
	sts curbuf,mpa 					;akt. Cursorpos lchen
scrlpc:
	st Y+,mpa						;ein Zeichen lschen
	dec mpb
	brne scrlpc
	rjmp scrlpa						;fertig

;Bildschirm lschen 
clrscr:
	ldi mpb,cpl				;eine Zeile
	dec mpc				;nchstes Mal die nchste Zeile
	mul mpb,mpc			;errechen Offset zum Lschen
	sts clrs, mpc

	ldi YH,HIGH(Charbuf) 	;Start Zeichenspeicher
	ldi YL, LOW(Charbuf)
	sts curph,YH			;wird neue Cursorposition
	sts curpl,YL			;nach dem lschen
	add YL,r0 				;Start + Offset
	adc YH,r1
	ldi mpa,' '				;Lschen = Leerzeichen eintragen
	ldi mpb,cpl				;eine Zeile lschen
clrlp:
	st Y+,mpa				;lschen
	dec mpb
	brne clrlp
	sts curbuf,mpa			;Cursorbuffer auch lschen
	;Cursor blinken lassen
	lds mpb, posc
	sbrc mpb, 3			;nur blinken lassen falls aktiv
	lds mpa,cur
	sbrs mpb, 3
	lds mpa, curbuf
	sbrc ticks,4			;ca. 2 Hz
	lds mpa,curbuf
	lds XH,curph
	lds XL,curpl
	st X,mpa				;auf akt. Cursorposition
	ret						;zurck zum INT-Handler

charwbz: rjmp charwb			;Zwischenstation

;nchstes Zeichen aus Input Buffer holen
;(wenn noch eins da ist)
charw:
	lds XH,curph			;Pointer in Charbuf
	lds XL,curpl			;schon mal vorsorglich holen
	cp ibupr,ibupw			;neues Zeichen im Input-Buffer ?
	breq charwbz			; Nein
	ldi YH,HIGH(inbuf)		;RSR232-Input Buffer
	ldi YL, LOW(inbuf)
	add YL,ibupr			;Bufferpointer-Read
	adc YH,NULL
	ld mpa,Y				;Bufferinhalt lesen
	inc ibupr				;Bufferpointer weitersetzen
	ldi mpb,InBufSize-1
	and ibupr,mpb			;Bufferpointer wraparound
	
	lds mpb, posc

	sbrc mpb, 0			;sind wir in einer X-Cursorpos-Sequenz ?
	rjmp charXPa			;  ja
	sbrc mpb, 1			;sind wir in einer Y-Cursorpos-Sequenz ?
	rjmp charYPa			;  ja
	sbrc mpb, 4			;sind wir in einer Curser-Sequenz?
	rjmp charCur			;  ja

	cpi mpa,0x01			;SOH? Ende
	breq char01z
	cpi mpa,0x02			;STX? Cur an Anfang
	breq char02z
	cpi mpa,0x03			;ETX? Cur an Ende
	breq char03z
	cpi mpa,0x04			;EOT? XPos ausgeben
	breq char04z
	cpi mpa,0x05			;ENQ? YPos ausgeben
	breq char05z
	cpi mpa,0x06			;ACK? Zeichen an Curpos ausgeben
	breq char06z
	cpi mpa,0x07			;BEL? POSC ausgeben
	breq char07z
	cpi mpa,0x08			;BS? Letztes Zeichen lschen
	breq char08z
	cpi mpa,0x09			;TAB? Tabulator
	breq char09x
	cpi mpa,0x0a			;LF? Nchste Zeile
	breq char0ax
	cpi mpa,0x0c			;FF? Screen leeren
	breq char0cx	
	cpi mpa,0x0d			;CR? Schreibkopfrcklauf
	breq char0dx
	cpi mpa,0x0e			;SO? Cur Y setzen
	breq charYPx
	cpi mpa,0x0f			;SI? Cur X setzen
	breq charXPx
	cpi mpa,0x10			;DLE? Cursor setzen
	breq char10z
	cpi mpa,0x11			;DC1? Cursor nach rechts
	breq char11z
	cpi mpa,0x12			;DC2? Cursor nach links
	breq char12z
	cpi mpa,0x13			;DC3? Cursor noch oben
	breq char13z
	cpi mpa,0x14			;DC4? Cursor nach unten
	breq char14z
	cpi mpa,0x15			;NAK? CursorShift an
	breq char15z
	cpi mpa,0x16			;SYN? CursorShift aus
	breq char16z
	cpi mpa,0x17			;ETB? Cursor an
	breq char17z
	cpi mpa,0x18			;CAN? Cursor aus
	breq char18z
	cpi mpa,0x19			;EM? Cursor ausgeben
	breq char19z

	rjmp sprung

char01z: rjmp char01			;Zwischenstation
char02z: rjmp char02			;Zwischenstation
char03z: rjmp char03			;Zwischenstation
char04z: rjmp char04			;Zwischenstation
char05z: rjmp char05			;Zwischenstation
char06z: rjmp char06			;Zwischenstation
char07z: rjmp char07			;Zwischenstation
char08z: rjmp char08			;Zwischenstation
char09x: rjmp char09			;Zwischenstation
char0ax: rjmp char0a			;Zwischenstation
char0cx: rjmp char0c			;Zwischenstation
char0dx: rjmp char0d			;Zwischenstation
charypx: rjmp charYP			;Zwischenstation
charxpx: rjmp charXP			;Zwischenstation
char10z: rjmp char10			;Zwischenstation
char11z: rjmp char11			;Zwischenstation
char12z: rjmp char12		 	;Zwischenstation
char13z: rjmp char13			;Zwischenstation
char14z: rjmp char14			;Zwischenstation
char15z: rjmp char15			;Zwischenstation
char16z: rjmp char16			;Zwischenstation
char17z: rjmp char17			;Zwischenstation
char18z: rjmp char18			;Zwischenstation
char19z: rjmp char19			;Zwischenstation
		
sprung:
	sbrc mpb, 2			;soll der Cursor weitergerckt werden - ja
	st X+,mpa				;nur Zeichen laden
	sbrs mpb, 2			; - nein
	st X,mpa				;ab in den Zeichenspeicher damit und nchste Stelle
	;sind wir em Bildschirmende ?
charwa:
	cpi XH,HIGH(Charbuf+cbsize)
	brlo charwc				; nein sicher nicht
	cpi XL,LOW(Charbuf+cbsize)
	brlo charwc				; nein
	;ja, dann mssen wir scrollen
	ldi mpa, txtl+1
	sts scroll,mpa		;txtl Zeilen scrollen + Leerzeile
	ldi XH,HIGH(Charbuf+CBSize-cpl)	;neue Cursorposition wird 
	ldi XL, LOW(Charbuf+CBSize-cpl) ;  Anfang letzte Zeile sein
charwc:
	ld mpa, X
	sts curbuf, mpa				;save Char auf Cursorposition
	;Cursor blinken lassen
charwb:
	lds mpb, posc
	sbrc mpb, 3			;nur blinken lassen falls aktiv
	lds mpa,cur				;--abwechselnd Cursor
	sbrs mpb, 3
	lds mpa, curbuf
	sbrc ticks,4			;ca. 2Hz Blinkfrequenz
	lds mpa, curbuf			;--oder aktuelles Zeichen
	st X,mpa				;in den Zeichenspeicher auf Cursorpos.
	sts curph,XH
	sts curpl,XL
	ret						;Zurck zum INT-Handler

;Cursorpositionierungssequenz X starten
charXP: 
	cbr mpb, 2
	cbr mpb, 16
	sbr mpb, 1
	sts posc, mpb
	rjmp charwb

;Cursorpositionierungssequenz Y starten
charYP: 
	cbr mpb, 1
	cbr mpb, 16
	sbr mpb, 2
	sts posc, mpb
	rjmp charwb

;LF Cursor eine Zeile tiefer setzen
char0a:
	lds mpa, curbuf
	st X, mpa				;Restore Char auf akt. Cursorpos
	cpi XH,HIGH(Charbuf+cbsize-cpl)	;schon in der letzten Zeile ?
	brlo char0aa			;  nein
	cpi XL,LOW(Charbuf+cbsize-cpl)
	brlo char0aa			;  nein
	ldi mpa, txtl+1
	sts scroll,mpa		;  ja, dann scrollen
	rjmp charwb
char0aa:
	adiw XH:XL,cpl			;akt.Position + ZeichenproZeile
								;  = eine Zeile tiefer
	rjmp charwa

;CR Cursor auf Anfang der Zeile setzen
char0d:
	lds mpa, curbuf
	st X, mpa				;Restore Char auf akt. Cursorpos
	sbiw XH:XL,Charbuf/2	;Berechne abs. Cursorpos im Buffer 
	sbiw XH:XL,Charbuf/2	;  also ohne Mem-Offset
	mov mpb,XH
	mov mpc,XL
	;berechne die atk. Cursor-Spalten Nr. 
char0da:
	sbiw XH:XL,cpl			;Primitive Methode um den Rest
	brcc char0da			;von Division / cpl zu erhalten
	subi XL,0-cpl			;egal, hier ist Zeit genug
	clr XH
	sub mpc,XL 				;Cursorspaltennr. von 
	sbci mpb,0				;Cursorposition abziehen
	mov XH,mpb				;ergibt Cursor auf Anfang Zeile
	mov XL,mpc
	adiw XH:XL,Charbuf/2	;Und den Speicher-Offset wieder dazu
	adiw XH:XL,Charbuf/2
	rjmp charwa				;fertig

;FF Bildschirm-Lschen auslsen
char0c:
	ldi mpa, txtl
	sts clrs,mpa			;alle Text-Zeilen lschen
	rjmp charwb

charYPa:
	cbr mpb, 2				;Ende Corsorpos. Sequenz
	sts posc, mpb
	cpi mpa,txtl			;neue Cursorposition > letzte Zeile
	brsh charwbx			;  ja -> ignorieren
	lds mpc, curbuf
	st X, mpc				;Restore Char auf Cursorpos
	;berechne die akt. Cursor-Spalten Nr. 
	sbiw XH:XL,Charbuf/2	;Berechne abs. Cursorpos im Buffer  
	sbiw XH:XL,Charbuf/2	;  ohne Mem-Offset
charYPb:
	sbiw XH:XL,cpl			;und nochmal die simple Division
	brcc charYPb			;weil sie so gut klappt
	subi XL,0-cpl
	clr XH
	ldi mpb,cpl				;Anzahl Zeichen pro Zeile
	mul mpa,mpb				;mal neue Zeilennr.
	add XL,r0				;plus Spaltennr.
	adc XH,r1
	adiw XH:XL,Charbuf/2	;und noch dem MEM-Offset
	adiw XH:XL,Charbuf/2 
	rjmp charwc				;setze neue Cursorpos.

;Cursor auf Spalte mpa setzen
charXPa:
	cbr mpb, 1				;Ende Corsorpos. Sequenz
	sts posc, mpb
	cpi mpa,cpl				;neue Cursorposition > cpl ?
	brsh charwbx			;  ja -> igorieren

	lds mpc, curbuf
	st X, mpc				;Restore Char auf aktueller Cursorpos
	;berechne die Startadresse der Zeile wo der Cursor drin steht 
	sbiw XH:XL,Charbuf/2	;Berechne abs. Cursorpos im Buffer
	sbiw XH:XL,Charbuf/2	;  ohne Mem-Offset
	mov mpb,XH				;sichern
	mov mpc,XL
charXPb:
	sbiw XH:XL,cpl			;Jaja, wieder die ALDI-Division
	brcc charXPb
	subi XL,0-cpl
	clr XH
	sub mpc,XL
	sbci mpb,0
	;neue X-Position auf Startadresse draufaddieren
	add mpc,mpa				;neue xpos
	adc mpb,NULL
	mov XH,mpb
	mov XL,mpc
	adiw XH:XL,Charbuf/2	;MEM-Offset wieder drauf
	adiw XH:XL,Charbuf/2
	rjmp charwc				;Cursor steht auf neuer Position

charwbx: rjmp charwb			;Zwischenstation
charwcx: rjmp charwc			;Zwischenstation

;DC1 Cursor nach rechts
char11:
	lds mpa, curbuf
	st X, mpa
	cpi XH,HIGH(Charbuf+cbsize-1)
	brlo char11x
	brne charwcx
	cpi XL, LOW(Charbuf+cbsize-1)
	brlo char11x
	rjmp charwc
char11x:
	adiw XH:XL,1
	rjmp charwc

;DC2 Cursor nach links
char12:
	lds mpa, curbuf
	st X, mpa
	cpi XH,HIGH(Charbuf+1)
	brlo charwcx
	brne char12x
	cpi XL, LOW(Charbuf+1)
	brlo charwcx
char12x:
	sbiw XH:XL,1
	rjmp charwc

;DC3 Cursor noch oben
char13:
	lds mpa, curbuf
	st X, mpa
	cpi XH,HIGH(Charbuf+cpl)
	brlo charwcx
	brne char13x
	cpi XL, LOW(Charbuf+cpl)
	brlo charwcx
char13x:
	sbiw XH:XL,cpl
	rjmp charwc

;DC4 Cursor nach unten
char14:
	lds mpa, curbuf
	st X, mpa
	cpi XH,HIGH(Charbuf+cbsize-cpl)
	brlo char14x
	brne charwcx
	cpi XL, LOW(Charbuf+cbsize-cpl)
	brlo char14x
	rjmp charwc
char14x:
	adiw XH:XL,cpl
	rjmp charwc

;Tabualtor
char09:
	lds mpa, curbuf
	st X, mpa
	cpi XH,HIGH(Charbuf+cbsize-4)
	brlo char09a
	brne charwcx
	cpi XL, LOW(Charbuf+cbsize-4)
	brlo char09a
	rjmp charwc
char09a:
	adiw XH:XL, 4
	andi XL, 0xFC
	rjmp charwc

;Backspace
char08:
	lds mpa, curbuf
	st X, mpa
	cpi XH,HIGH(Charbuf+1)
	brlo charwcx
	brne char08x
	cpi XL, LOW(Charbuf+1)
	brlo charwcx
char08x:
	sbiw XH:XL,1
	ldi mpb, ' '
	st X,mpb
	rjmp charwc

;Char ausgabe
char06:
	lds outdata, curbuf
	rjmp charwb

;STX
char02:
	lds mpa, curbuf
	st X, mpa
	ldi XH,HIGH(Charbuf)
	ldi XL, LOW(Charbuf)
	rjmp charwc

;ETX
char03:
	lds mpa, curbuf
	st X, mpa
	ldi XH,HIGH(Charbuf+cbsize-1)
	ldi XL, LOW(Charbuf+cbsize-1)
	rjmp charwc

;Xpos
char04:
	mov mpb, XH
	mov mpc, XL
	sbiw XH:XL,Charbuf/2
	sbiw XH:XL,Charbuf/2
char04x:
	sbiw XH:XL, cpl
	brcc char04x
	subi XL, 0-cpl
	mov outdata, XL
	mov XH, mpb
	mov XL, mpc
	rjmp charwb

;Ypos
char05:
	mov mpb, XH
	mov mpc, XL
	ldi mpa, 0
	sbiw XH:XL,Charbuf/2
	sbiw XH:XL,Charbuf/2
char05x:
	inc mpa
	sbiw XH:XL, cpl
	brcc char05x
	dec mpa
	mov outdata, mpa
	mov XH, mpb
	mov XL, mpc
	rjmp charwb

;Cursorshift an
char15:
	sbr mpb, 4
	sts posc, mpb
	rjmp charwb

;Cursorshift aus
char16:
	cbr mpb, 4
	sts posc, mpb
	rjmp charwb

;Cursor an
char17:
	sbr mpb, 8
	sts posc, mpb
	rjmp charwb

;Cursor aus
char18:
	cbr mpb, 8
	sts posc, mpb
	rjmp charwb

;Ende
char01:
	ldi mpa, cpl-1
	rjmp CharXPa

;Settings
char07:
	lds outdata, posc
	rjmp charwb

;Cursersequentz starten
char10:
	cbr mpb, 2
	cbr mpb, 1
	sbr mpb, 16
	sts posc, mpb
	rjmp charwb

;Cursor setzen
charCur:
	cbr mpb, 16
	sts posc, mpb
	sts cur, mpa
	rjmp charwb

;Cursor ausgeben
char19:
	lds mpa, cur
	mov outdata, mpa
	rjmp charwb

 ;Charactergenerator
.ORG 0xC00
.include "zeichensatz.txt"
;ENDE
